#![allow(unused)]

use core::{cell::Cell, convert::Infallible};

use arduino_hal::{
    delay_us,
    pac::{EXINT, TC0},
    port::{
        mode::{Input, OpenDrain, PullUp},
        Pin, PinMode, PinOps,
    },
};
use avr_device::{asm::delay_cycles, interrupt};
use embedded_hal::digital::{InputPin, OutputPin};

use super::static_pins_impl::StaticPin;
use heapless::Vec;

pub struct HalfDuplexSerial<P> {
    _pin: Pin<Input<PullUp>, P>,
}

#[thread_local]
static SERIAL_BUFFER: Cell<Vec<u8, 20>> = Cell::new(Vec::new());

const SERIAL_TRANSMIT_DELAY_US: u32 = 4;
const LOGIC_STATE_CHANGE_DELAY_CYCLES: u32 = 32;
const POLL_DELAY: u32 = 12;

const MSB: u8 = 0x80;

pub type PollResult = Result<(), PollError>;

pub enum PollError {
    NotFound,
    NotReady,
}

pub type ReadByteResult = Result<u8, ReadByteError>;

pub enum ReadByteError {
    CorruptedData,
    Other,
}

pub trait SoftSerial<P>
where
    P: PinOps + StaticPin,
{
    fn poll(&self) -> PollResult;
    fn response(&self);

    fn write_byte(&self, data: u8);
    fn read_byte(&self) -> ReadByteResult;

    fn write_bytes(&self, transmit_data: &[u8]);
    fn read_bytes(&self, recieve_buffer: &mut [u8]);
}

impl<P> HalfDuplexSerial<P>
where
    P: PinOps + StaticPin,
{
    #[inline]
    pub fn new(_pin: Pin<Input<PullUp>, P>, exint: &EXINT) -> Self {
        exint.eimsk.write(|w| w.bits(1 << 3));
        exint.eicra.write(|w| unsafe { w.bits(!(1 << 6 | 1 << 7)) });
        Self { _pin }
    }
}

impl<P> SoftSerial<P> for HalfDuplexSerial<P>
where
    P: PinOps + StaticPin,
{
    fn poll(&self) -> PollResult {
        let pin_pos = 1u8 << P::PIN_NUM;

        P::into_input();

        if P::read() & pin_pos == 0 {
            return PollResult::Err(PollError::NotFound);
        }

        delay_us(3);

        if P::read() & pin_pos != 0 {
            return PollResult::Err(PollError::NotReady);
        }

        P::into_pullup_input();

        PollResult::Ok(())
    }

    fn response(&self) {
        delay_us(1);

        P::into_output();

        delay_us(4);

        P::into_pullup_input();
    }

    fn write_byte(&self, data: u8) {
        let mut data = data;
        let mut parity_bit = 0;
        let pin_pos = 1u8 << P::PIN_NUM;

        P::write(P::read() & !pin_pos);
        P::into_output();

        for _ in 0..8 {
            // Since some Atmega MCUs don't have a opendrain pin mode, by switching pins mode we
            // can simulate opendrain mode
            //
            // Opendrain pin mode simulation example:
            //
            // P::into_output() -> set low;
            // P::into_input() -> set high;

            if data & MSB == 0 {
                P::into_output();
            } else {
                P::into_input();
                parity_bit ^= 1;
            }

            data <<= 1;
            delay_us(SERIAL_TRANSMIT_DELAY_US);
        }

        // Hamming code and CRC are very weightful and slow, so I use simple parity check

        if parity_bit == 1 {
            P::into_input();
        }

        delay_us(SERIAL_TRANSMIT_DELAY_US);

        P::into_pullup_input();
    }

    fn read_byte(&self) -> ReadByteResult {
        let mut data = 0u8;
        let mut parity_bit = 0u8;
        let pin_pos = 1u8 << P::PIN_NUM;

        for _ in 0..8 {
            delay_cycles(LOGIC_STATE_CHANGE_DELAY_CYCLES);

            if P::read() & pin_pos != 0 {
                data |= 1;
                parity_bit ^= 1;
            }

            data <<= 1;

            delay_cycles(LOGIC_STATE_CHANGE_DELAY_CYCLES);
        }

        delay_cycles(LOGIC_STATE_CHANGE_DELAY_CYCLES);

        let bit_parity_check = P::read() & pin_pos != parity_bit;

        delay_cycles(LOGIC_STATE_CHANGE_DELAY_CYCLES);

        if !bit_parity_check {
            return Err(ReadByteError::CorruptedData);
        }

        Ok(data)
    }

    fn write_bytes(&self, transmit_data: &[u8]) {
        for byte in transmit_data {
            self.write_byte(*byte);
        }
    }

    fn read_bytes(&self, recieve_buffer: &mut [u8]) {
        for byte in recieve_buffer {
            match self.read_byte() {
                Ok(data) => *byte = data,
                Err(_) => continue,
            }
        }
    }
}
