#![no_std]
#![no_main]
#![feature(panic_info_message)]
#![feature(abi_avr_interrupt)]
#![feature(asm_experimental_arch)]
#![feature(thread_local)]

#[macro_use]
extern crate static_pins;

use arduino_hal::{
    delay_ms, delay_us,
    hal::{
        pins,
        port::{PD2, PD3, PD5},
    },
    port::{mode::Output, Pin},
    Peripherals,
};

mod half_duplex_serial;
mod keyboard;
mod panic_handler;
mod static_pins_impl;

use avr_device::interrupt::{enable, free};
use half_duplex_serial::{HalfDuplexSerial, PollError, SoftSerial};
use keyboard::*;

#[allow(unused)]
use panic_handler::*;

// static mut SERIAL_BUFFER: Vec<u8, 5> = Vec::new();

pub static mut SERIAL: Option<HalfDuplexSerial<PD3>> = None;

static mut RX_LED: Option<Pin<Output, PD5>> = None;
static mut POLL_INDICATOR: Option<Pin<Output, PD2>> = None;

// const MODIFIER_MASK: u8 = 0xE0;

#[arduino_hal::entry]
fn main() -> ! {
    let dp = Peripherals::take().unwrap();

    let pins = pins!(dp);

    let mut matrix = KeyboardMatrix::new(
        pins_array!("cols", pins, pd0, pd4),
        pins_array!("rows", pins, pd7),
    );

    let mut rx_led = pins.pd5.into_output();
    let poll_indicator = pins.pd2.into_output();
    // let mut delay: Delay<MHz16> = Delay::new();

    pins.pb0.into_output().set_low();

    rx_led.set_low();

    let data_line = HalfDuplexSerial::new(pins.pd3.into_pull_up_input(), &dp.EXINT);

    unsafe {
        POLL_INDICATOR = Some(poll_indicator);
        SERIAL = Some(data_line);
        RX_LED = Some(rx_led);
        enable();
        loop {
            let scan = matrix.scan();
            if scan[0].is_some() || scan[1].is_some() {
                if let Some(data_line) = SERIAL.as_mut() {
                    match data_line.poll() {
                        Ok(_) => {
                            if let Some(rx_led) = RX_LED.as_mut() {
                                rx_led.toggle();
                            }
                        }
                        Err(poll_error) => match poll_error {
                            PollError::NotFound => {
                                // panic!("1");
                            }
                            PollError::NotReady => {
                                // panic!("2");
                            }
                        },
                    }
                }
            }
        }
    }
}

#[avr_device::interrupt(atmega32u4)]
unsafe fn INT3() {
    free(|_| {
        if let Some(serial) = SERIAL.as_ref() {
            serial.response();
        }
        if let Some(poll_status) = POLL_INDICATOR.as_mut() {
            poll_status.toggle();
        }
        // if let Some(rx_led) = RX_LED.as_mut() {
        //     rx_led.toggle();
        // }
    })
}
